/*
 * Copyright (C) 2010 Nokia Corporation.
 *
 * Contact: Marius Vollmer <marius.vollmer@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

/* This program will write supplements for *.desktop files into well
   known places.  These supplements reflect the current state of the
   packages that own the primary desktop files.

   The supplements only reflect the 'static' state of the packages;
   whether they are installed; whether an update is available, etc.
   They do not contain dynamic information, such as whether a package
   is currently being downloaded and how far along this is.  (On
   Harmattan, this dynamic information is available as D-Bus signals
   emitted by the package manager daemon).

   This program will be called by apt at three different times:

   - After updating the package lists (apt-get update on the
     command line).

   - Before starting the download phase that precedes a dpkg
     operation.

   - After invoking dpkg.

   At all times, apt-desktop will iterate through all "user visible"
   packages, and write all .desktop files that it can, either based on
   the MeeGo-Desktop-Entry etc meta data fields, or by guessing the
   information from the list of files in the package.

   This program also maintains a little database with a mapping from
   package name to desktop file names.  This database can be queried
   by running apt-desktop on the command line (the "list" and "get"
   sub-commands), or by reading the symlinks in a directory:

       /var/lib/apt-desktop/entries/PACKAGE

   points to the desktopfile for PACKAGE.  The target of the symlink
   is not guaranteed to exist.  Sometimes it is known which desktop
   file is in a package before that package is actually installed.
*/

#define _ATFILE_SOURCE

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <libgen.h>

#include <apt-pkg/init.h>
#include <apt-pkg/error.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/tagfile.h>

#include <map>
#include <string>

using namespace std;

#define DEFAULT_DB_PATH    "/var/lib/apt-desktop/entries"
#define DEFAULT_BASE       "/usr/share/applications"
#define DEFAULT_STATE_DIR  "/var/lib/apt-desktop"

/* UTILITIES
 */

int flag_debug = 0;
int flag_verbose = 0;
int flag_all_packages = 0;

void
dbg (const char *fmt, ...)
{
  if (flag_debug)
    {
      va_list ap;
      va_start (ap, fmt);
      vfprintf (stdout, fmt, ap);
      va_end (ap);
    }
}

void
exitf (int code, const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  exit (1);
}

char *
must_strdup (const char *str)
{
  char *dup = strdup (str);
  if (dup == NULL)
    exitf (1, "Out of memory.\n");
  return dup;
}

char *
must_strndup (const char *str, int n)
{
  char *dup = strndup (str, n);
  if (dup == NULL)
    exitf (1, "Out of memory.\n");
  return dup;
}

void *
must_realloc (void *mem, size_t n)
{
  void *remem = realloc (mem, n);
  if (remem == NULL && n != 0)
    exitf (1, "Out of memory.\n");
  return remem;
}

void
must_chdir (const char *path)
{
  if (chdir (path) < 0)
    {
      if (access (path, F_OK) < 0 && mkdir (path, 0755) < 0)
        exitf (1, "Can't create directory %s: %m\n", path);
      if (chdir (path) < 0)
        exitf (1, "Can't change current directory to %s: %m\n", path);
    }
}

void
must_symlinkat (const char *to, int fd, const char *from)
{
  if (symlinkat (to, fd, from) < 0)
    exitf (1, "Can't symlink %s to %s: %m\n", from, to);
}

void
must_unlinkat (int fd, const char *path)
{                                                     
  if (unlinkat (fd, path, 0) < 0)
    exitf (1, "Can't unlink %s: %m\n", path);
}

void
must_disappear (const char *path)
{                                                     
  if (unlink (path) < 0 && errno != ENOENT)
    exitf (1, "Can't unlink %s: %m\n", path);
}

int
must_openat (int dfd, const char *path, int flags, mode_t mode)
{                                                     
  int fd = openat (dfd, path, flags, mode);
  if (fd < 0)
    exitf (1, "Can't open %s: %m\n", path);
  return fd;
}

void
must_close (int fd)
{                                                     
  if (close (fd) < 0)
    exitf (1, "Can't close: %m\n");
}

void
must_fsync (int fd)
{                                                     
  if (fsync (fd) < 0)
    exitf (1, "Can't sync: %m\n");
}

void
must_read (int fd, char *buf, int n)
{          
  while (n > 0)
    {
      int r = read (fd, buf, n);
      if (r < 0)
        exitf (1, "Can't read: %m\n");
      if (r == 0)
        exitf (1, "Unexpected end of file.\n");
      n -= r;
      buf += r;
    }
}

void
must_write (int fd, const char *buf, int n)
{          
  while (n > 0)
    {
      int r = write (fd, buf, n);
      if (r < 0)
        exitf (1, "Can't write: %m\n");
      n -= r;
      buf += r;
    }
}

void
must_rename (const char *from, const char *to)
{                                                     
  if (rename (from, to) < 0)
    exitf (1, "Can't rename %s to %s: %m\n", from, to);
}

void
must_renameat (int from_fd, const char *from, int to_fd, const char *to)
{                                                     
  if (renameat (from_fd, from, to_fd, to) < 0)
    exitf (1, "Can't rename %s to %s: %m\n", from, to);
}

FILE *
must_popenf (const char *mode, const char *fmt, ...)
{
  char *cmd;
  va_list ap;
  va_start (ap, fmt);
  vasprintf (&cmd, fmt, ap);
  va_end (ap);

  FILE *p = popen (cmd, mode);
  if (p == NULL)
    exitf (1, "Can't run '%s': %m", cmd);
  
  free (cmd);
  return p;
}

void
must_pclose (FILE *p, const char *cmd)
{
  if (pclose (p) != 0)
    exitf (1, "Command %s terminated unsuccessfully.\n", cmd);
}

DIR *
must_fdopendir (int fd)
{
  DIR *d = fdopendir (fd);
  if (d == NULL)
    exitf (1, "Can't list directory: %m\n");

  return d;
}

void
must_closedir (DIR *d)
{
  if (closedir (d) != 0)
    exitf (1, "Can't stop listen directory........ %m\n");
}

char *
xreadlinkat (int fd, const char *path)
{
  int size = 100;
  char *buffer = NULL;

  while (1)
    {
      buffer = (char *) must_realloc (buffer, size);
      int nchars = readlinkat (fd, path, buffer, size);
      if (nchars < 0)
        {
          free (buffer);
          return NULL;
        }
      if (nchars < size)
        {
          buffer[nchars] = '\0';
          return buffer;
        }
      size *= 2;
    }
}

void
update_if_changed (int dfd, const char *path, const char *entry)
{
  struct stat buf;
  int n = strlen (entry);

  // Return early when the old content is already the same as ENTRY.

  if (fstatat (dfd, path, &buf, 0) == 0)
    {
      if (buf.st_size == n)
        {
          char *old = new char[n];
          int fd = openat (dfd, path, O_RDONLY);
          if (fd >= 0)
            {
              if (read (fd, old, n) == n)
                {
                  if (memcmp (old, entry, n) == 0)
                    {
                      free (old);
                      dbg ("File %s needs no change.\n", path);
                      return;
                    }
                }
            }
          free (old);
        }
    }

  // We need to change the file.
  
  char *path_tmp;
  asprintf (&path_tmp, "%s.tmp", path);
  int fd = must_openat (dfd, path_tmp, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  must_write (fd, entry, n);
  must_fsync (fd);
  must_close (fd);
  must_renameat (dfd, path_tmp, dfd, path);
  free (path_tmp);
}

void
must_mkdir (const char *path)
{
  if (mkdir (path, 0755) < 0)
    exitf (1, "Can't create directory %s: %m\n", path);
}

void
ensure_directory (const char *path)
{
  struct stat buf;
  char *dup = must_strdup (path);
  char *dir = dirname (dup);
  if (stat (dir, &buf) == -1)
    {
      ensure_directory (dir);
      must_mkdir (dir);
    }
  free (dup);
}

/* OUR STATE

   Our state is just a big string.  We load a old version at the
   beginning and then construct a completely new string while working.
   That new string is saved as the state for the next invocation of
   apt-desktop.
 */

char *old_state_file_name;
char *new_state_file_name;

char *old_state;
FILE *new_state_file;

void
init_state ()
{
  struct stat buf;

  const char *state_dir = getenv ("APT_DESKTOP_STATE_DIR");
  if (state_dir == NULL)
    state_dir = DEFAULT_STATE_DIR;

  asprintf (&old_state_file_name, "%s/state", state_dir);
      
  old_state = NULL;
  if (stat (old_state_file_name, &buf) >= 0)
    {
      int fd = must_openat (AT_FDCWD, old_state_file_name, O_RDONLY, 0);
      old_state = new char[buf.st_size+1];
      must_read (fd, old_state, buf.st_size);
      old_state[buf.st_size] = '\0';
      must_close (fd);
    }

  asprintf (&new_state_file_name, "%s.new", old_state_file_name);
  new_state_file = fopen (new_state_file_name, "w");
  if (new_state_file == NULL)
    exitf (1, "Can't open %s: %m", new_state_file_name);
}

char *
get_state (const char *type, const char *key)
{
  char *p = old_state;
  if (p == NULL)
    return NULL;

  while (1)
    {
      char *p_end = strchr (p, '\n');
      if (p_end == NULL)
        break;

      char *p_type = p;
      char *p_type_end = strchr (p_type, ' ');
      if (p_type_end == NULL)
        break;

      if (strlen (type) == p_type_end - p_type
          && strncmp (type, p_type, p_type_end - p_type) == 0)
        {
          char *p_key = p_type_end + 1;
          char *p_key_end = strchr (p_key, ' ');
          if (p_key_end == NULL)
            break;

          if (strlen (key) == p_key_end - p_key
              && strncmp (key, p_key, p_key_end - p_key) == 0)
            {
              char *val = p_key_end + 1;
              char *val_end = strchr (val, '\n');
              if (val_end == NULL)
                break;
              
              return must_strndup (val, val_end - val);
            }
        }

      p = p_end + 1;
    }

  return NULL;
}

void
append_new_state (const char *type, const char *key, const char *valfmt, ...)
{
  fprintf (new_state_file, "%s %s ", type, key);
  va_list ap;
  va_start (ap, valfmt);
  vfprintf (new_state_file, valfmt, ap);
  va_end (ap);
  fprintf (new_state_file, "\n");
}

void
save_new_state ()
{
  if (fflush (new_state_file) != 0
      || fsync (fileno (new_state_file)) < 0
      || fclose (new_state_file) != 0)
    exitf (1, "Error closing %s: %m\n", new_state_file_name);

  must_rename (new_state_file_name, old_state_file_name);
}

/* EXPLICITLY INTERESTING PACKAGES

   We write desktop entries for all 'interesting' packages.  A package
   is either implicitly interesting (when it is installed and user
   visible) or explicitly interesting (when we have been told
   explicitly that it is interesting, via the set-interesing command).

   For each explicitly interesting package, we also store some extra
   information in the form of an integer.  If that integer is zero,
   the package is not explicitly interesting.  Other values are used
   to record whether the package had an error, etc.

   We keep a list of explicitly interesting packages in our state
   directory, together with the extra information.
*/

enum {
  interest_normal = 1,
  interest_had_error = 2
};

map<string,int> interesting_packages;
bool interesting_packages_dirty = false;

void
init_interesting_packages ()
{
  const char *state_dir = getenv ("APT_DESKTOP_STATE_DIR");
  if (state_dir == NULL)
    state_dir = DEFAULT_STATE_DIR;
  
  char *name;
  asprintf (&name, "%s/interesting", state_dir);

  FILE *f = fopen (name, "r");
  if (f)
    {
      char *line = NULL;
      size_t len = 0;
      ssize_t n;
      while ((n = getline (&line, &len, f)) != -1)
	{
	  if (n > 0 && line[n-1] == '\n')
	    line[n-1] = '\0';

          char *p = line;
          char *e = strchr (p, ' ');
          int reason = interest_normal;
          if (e)
            {
              *e++ = '\0';
              reason = atoi (e);
            }

          interesting_packages[p] = reason;
	}
      interesting_packages_dirty = false;

      free (line);
      fclose (f);
    }
  else if (errno != ENOENT)
    exitf (1, "Can't open %s: %m", name);

  free (name);
}

void
save_interesting_packages ()
{
  if (!interesting_packages_dirty)
    return;

  const char *state_dir = getenv ("APT_DESKTOP_STATE_DIR");
  if (state_dir == NULL)
    state_dir = DEFAULT_STATE_DIR;
  
  char *name;
  asprintf (&name, "%s/interesting.new", state_dir);

  FILE *f = fopen (name, "w");
  if (f == NULL)
    exitf (1, "Can't open %s: %m", name);

  for (map<string,int>::iterator it = interesting_packages.begin();
       it != interesting_packages.end(); it++)
    {
      if (it->second)
        fprintf (f, "%s %d\n", it->first.c_str(), it->second);
    }

  if (fflush (f) != 0
      || fsync (fileno (f)) < 0
      || fclose (f) != 0)
    exitf (1, "Error closing %s: %m\n", name);

  char *real_name;
  asprintf (&real_name, "%s/interesting", state_dir);

  must_rename (name, real_name);

  free (name);
  free (real_name);
}

void
remember_interesting_package (const char *package, int reason)
{
  if (interesting_packages[package] != reason)
    {
      interesting_packages[package] = reason;
      interesting_packages_dirty = true;
    }
}

void
forget_interesting_package (const char *package)
{
  remember_interesting_package (package, 0);
}

int
explicitly_interesting (const char *package)
{
  return interesting_packages[package];
}

/* We used to have the concept of 'temporarily' explicitly interesting
   packages.  Those packages would be removed from the list of
   explicitly interesting packages as soon as they become implicitly
   interesting.

   We don't really have that concept anymore, but we keep the
   machinery around, for reference.
*/
bool
is_temporarily_explictly_interesting (const char *package)
{
  return false;
}

/* LIBAPT-PKG STATE
 */

pkgCacheFile *apt_cache_ptr;
#define       apt_cache (*apt_cache_ptr)

class SilentProgress : public OpProgress
{
  virtual void
  Update ()
  {
    // cerr << Percent << "\r";
  }
};

void
init_apt ()
{
  static bool done = false;

  if (!done)
    {
      done = true;

      if (pkgInitConfig (*_config) == false ||
          pkgInitSystem (*_config, _system) == false)
        _error->DumpErrors ();
    }

  apt_cache_ptr = new pkgCacheFile;

  SilentProgress progress;
  if (!apt_cache.Open (progress, false))
    {
      _error->DumpErrors ();
      exit (1);
    }
}

pkgCache::VerIterator
candidate (pkgCache::PkgIterator &pkg)
{
  pkgDepCache &cache = *apt_cache;
  return cache[pkg].CandidateVerIter(cache);
}

pkgCache::VerIterator
installed (pkgCache::PkgIterator &pkg)
{
  return pkg.CurrentVer();
}

bool
package_broken (pkgCache::PkgIterator &pkg)
{
  return pkg->Flags & pkgCache::Flag::Broken;
}

bool
package_depended_on_by_installed_system_update (pkgCache::VerIterator &ver, int level = 0)
{
  if (ver->Flags & pkgCache::Flag::OSMetaPackage)
    return true;

  /* Cheap ass disaster avoidance...
   */
  if (level > 10)
    return false;

  for (pkgCache::DepIterator rev = ver.ParentPkg().RevDependsList(); rev.end() == false; rev++)
    {
      pkgCache::PkgIterator r_pkg = rev.ParentPkg();
      pkgCache::VerIterator r_ver = rev.ParentVer();
      
      if ((rev->Type == pkgCache::Dep::PreDepends
           || rev->Type == pkgCache::Dep::Depends)
          && r_pkg.CurrentVer() == r_ver
          && package_depended_on_by_installed_system_update (r_ver, level+1))
        return true;
    }

  return false;
}

bool
package_removable (pkgCache::VerIterator &ver)
{
  return !package_depended_on_by_installed_system_update (ver);
}

int
should_consider_package (pkgCache::PkgIterator &pkg)
{
  return (!pkg.end()
          && (flag_all_packages || pkg->Flags & pkgCache::Flag::Visible)
          && !(pkg->Flags & pkgCache::Flag::OSMetaPackage));
}

struct package_record {
  pkgRecords Recs;
  pkgRecords::Parser &P;
  pkgTagSection section;
  bool valid;

  package_record (const pkgCache::VerIterator &ver);

  bool has (const char *tag);
  string get_string (const char *tag);
  char *get (const char *tag);
  int get_int (const char *tag, int def);
};

package_record::package_record (const pkgCache::VerIterator &ver)
  : Recs (*apt_cache),
    P (Recs.Lookup (ver.FileList ()))
{
  const char *start, *stop;
  P.GetRec (start, stop);

  /* NOTE: pkTagSection::Scan only succeeds when the record ends in
           two newlines, but pkgRecords::Parser::GetRec does not
           include the second newline in its returned region.
           However, that second newline is always there, so we just
           pass one more character to Scan.
  */
  
  valid = section.Scan (start, stop-start+1);
}

bool
package_record::has (const char *tag)
{
  unsigned pos;
  return valid && section.Find (tag, pos);
}

string
package_record::get_string (const char *tag)
{
  if (valid)
    return section.FindS (tag);
  else
    return string ("");
}
 
int
all_white_space (const char *text)
{
  while (*text)
    if (!isspace (*text++))
      return 0;
  return 1;
}

char *
package_record::get (const char *tag)
{
  if (!valid)
    return NULL;

  string res = get_string (tag);
  if (all_white_space (res.c_str ()))
    return NULL;
  else
    return must_strdup (res.c_str());
}

int
package_record::get_int (const char *tag, int def)
{
  if (!valid)
    return def;

  return section.FindI (tag, def);
}

/* UPDATING THE MAPPING DATABASE
 */

const char *db_path;
int db_fd;

void
init_db ()
{
  if (db_path == NULL)
    {
      db_path = getenv ("APT_DESKTOP_DB_PATH");
      if (db_path == NULL)
        db_path = DEFAULT_DB_PATH;
      
      if (access (db_path, F_OK) < 0 && mkdir (db_path, 0755) < 0)
        exitf (1, "Can't create directory %s: %m\n", db_path);
      db_fd = must_openat (AT_FDCWD, db_path, O_RDONLY, 0);
    }
}

void export_entry (const char *package, const char *path);

void
record_mapping (const char *package, const char *path)
{
  struct stat buf;
  char *full_path = (char *)path;

  dbg ("Found mapping %s -> %s.\n", package, path);

  if (strchr (full_path, '.') == NULL)
    asprintf (&full_path, "%s.desktop", full_path);
      
  if (full_path[0] != '/')
    asprintf (&full_path, "%s/%s", DEFAULT_BASE, full_path);

  if (fstatat (db_fd, package, &buf, AT_SYMLINK_NOFOLLOW) == 0)
    {
      if (S_ISLNK (buf.st_mode))
        {
          char *old_path = xreadlinkat (db_fd, package);
          if (strcmp (old_path, full_path) == 0)
            {
              free (old_path);
              goto export_it;
            }
          free (old_path);
        }
      must_unlinkat (db_fd, package);
    }

  must_symlinkat (full_path, db_fd, package);

 export_it:
  export_entry (package, full_path);

  if (full_path != path)
    free (full_path);
}

bool
update_version_from_content (const pkgCache::PkgIterator &pkg, const pkgCache::VerIterator &ver)
{
  bool found = false;

  if (!ver.end())
    {
      char *adminoption;
      char *line = NULL;
      size_t len = 0;
      ssize_t n;

      dbg ("Scanning %s %s.\n", pkg.Name(), ver.VerStr());
                
      if (getenv ("APT_DESKTOP_DPKG_ADMINDIR"))
        asprintf (&adminoption, "--admindir '%s' ", getenv ("APT_DESKTOP_DPKG_ADMINDIR"));
      else
        adminoption = must_strdup ("");

      regex_t pat1, pat2;
      regcomp (&pat1, "/usr/share/applications/[^/]\\+.desktop", 0);
      regcomp (&pat2, "/usr/share/meegotouch/applicationextensions/[^/]\\+.desktop", 0);

      FILE *p = must_popenf ("r", "dpkg-query %s-L %s", adminoption, pkg.Name());
      while ((n = getline (&line, &len, p)) != -1)
	{
	  if (n > 0 && line[n-1] == '\n')
	    line[n-1] = '\0';

          if (regexec (&pat1, line, 0, NULL, 0) == 0
              || regexec (&pat2, line, 0, NULL, 0) == 0)
            {
              record_mapping (pkg.Name(), line);
              found = true;
            }
	}
      free (line);
      must_pclose (p, "dpkg-query");

      free (adminoption);
    }

  return found;
}

bool
update_version (const pkgCache::PkgIterator &pkg, const pkgCache::VerIterator &ver)
{
  if (!ver.end())
    {
      package_record rec(ver);
      char *path = rec.get ("Meego-Desktop-Entry-Filename");
      if (path)
        {
          record_mapping (pkg.Name(), path);
          return true;
        }
      if (rec.has ("MeeGo-Desktop-Entry"))
        {
          record_mapping (pkg.Name(), pkg.Name());
          return true;
        }
    }
  
  return false;
}

char *
entry_path (const char *package)
{
  struct stat buf;
  char *to;

  if (fstatat (db_fd, package, &buf, AT_SYMLINK_NOFOLLOW) == 0)
    to = xreadlinkat (db_fd, package);
  else
    to = NULL;

  return to;
}

char *
extra_path (const char *path)
{
  char *p = strdup (path);
  char *pp;

  char *base = basename (p);
  asprintf (&pp, "%.*sinstaller-extra/%s", base-p, p, base);
  free (p);
  p = pp;

  const char *root = getenv ("APT_DESKTOP_EXPORT_ROOT");
  if (root)
    {
      char *pp;
      asprintf (&pp, "%s/%s", root, p);
      free (p);
      p = pp;
    }
  
  return p;
}

void
remove_stale_mappings ()
{
  struct dirent *ep;
  DIR *d = must_fdopendir (db_fd);
  while (ep = readdir (d))
    {
      if (strcmp (ep->d_name, ".") == 0
          || strcmp (ep->d_name, "..") == 0)
        continue;

      pkgCache::PkgIterator pkg = apt_cache->FindPkg (ep->d_name);
      if (!should_consider_package (pkg)
          && !explicitly_interesting (ep->d_name))
        {
          char *path = xreadlinkat (db_fd, ep->d_name);
          char *extra = extra_path (path);

          dbg ("Removing %s and %s\n", ep->d_name, extra);
          must_disappear (extra);
          must_unlinkat (db_fd, ep->d_name);

          free (extra);
          free (path);
        }
    }
  must_closedir (d);
}

/* CREATING ENTRIES
 */

void
decode_multiline_in_place (char *field)
{
  if (field == NULL)
    return;

  /* The decoding is this:

     - the first space of a line is removed.

     - if after that a line consists solely of a '.', that dot is
       removed.
  */

  char *src = field, *dst = field;
  char *bol = src;

  while (*src)
    {
      if (src == bol && *src == ' ')
	{
	  src++;
	  continue;
	}

      if (*src == '\n')
	{
	  if (bol+2 == src && bol[0] == ' ' && bol[1] == '.')
	    dst--;
	  bol = src + 1;
	}

      *dst++ = *src++;
    }

  *dst = '\0';
}

char *
create_entry (const char *package)
{
  pkgCache::PkgIterator pkg = apt_cache->FindPkg (package);
  pkgCache::VerIterator cand;
  pkgCache::VerIterator inst;

  if (!pkg.end())
    {
      cand = candidate (pkg);
      inst = installed (pkg);
    }

  const char *state;
  const char *update_available = NULL;
  const char *had_error = NULL;
  const char *removable = NULL;

  int interest_reason = explicitly_interesting (package);
  
  if (interest_reason == interest_had_error)
    had_error = "true";

  if (!inst.end())
    {
      if (package_broken (pkg))
        state = "broken";
      else 
        state = "installed";

      if (!cand.end() && cand != inst)
        update_available = "true";
    }
  else if (interest_reason != 0)
    {
      if (!cand.end())
        state = "installable";
      else
        state = "missing";
    }
  else
    state = NULL;
  
  if (state == NULL)
    return NULL;

  if (!inst.end() && !package_removable (inst))
    removable = "false";

  char *head = NULL;
  if (!cand.end())
    {
      package_record rec(cand);
      head = rec.get ("MeeGo-Desktop-Entry");
      if (head)
        decode_multiline_in_place (head);
    }

  if (head == NULL)
    head = must_strdup ("[Desktop Entry]\nType=Application");

  // XXX - merge the MeeGo-Desktop-Entry and our fields
  //       properly, don't assume we can just dump
  //       MeeGo-Desktop-Entry in front.
  
  char *update_available_field = NULL;
  if (update_available)
    asprintf (&update_available_field, "PackageUpdateAvailable=%s\n", update_available);
  else
    update_available_field = must_strdup ("");

  char *had_error_field = NULL;
  if (had_error)
    asprintf (&had_error_field, "PackageHadError=%s\n", had_error);
  else
    had_error_field = must_strdup ("");

  char *removable_field = NULL;
  if (removable)
    asprintf (&removable_field, "PackageRemovable=%s\n", removable);
  else
    removable_field = must_strdup ("");

  char *result;
  asprintf (&result, "%s\n\n[X-MeeGo]\nPackage=%s\nPackageState=%s\n%s%s%s",
            head, pkg.Name(), state,
            update_available_field, had_error_field, removable_field);

  free (head);
  free (update_available_field);
  free (had_error_field);
  
  return result;
}

void
export_entry (const char *package, const char *path)
{
  char *p = (char *)path;
  if (p == NULL)
    {
      p = entry_path (package);
      if (p == NULL)
        return;
    }
    
  
  char *extra = extra_path (p);
  char *entry = create_entry (package);

  if (entry)
    {
      dbg ("Writing %s for %s\n", extra, package);
      ensure_directory (extra);
      update_if_changed (AT_FDCWD, extra, entry);
    }
  else
    {
      dbg ("Removing %s for %s\n", extra, package);
      must_disappear (extra);
    }

  if (p != path)
    free (p);
  free (extra);
  free (entry);
}

/* MAIN 
 */

void
usage ()
{
  fprintf (stderr, "Usage: apt-desktop [-Dva] update\n");
  fprintf (stderr, "       apt-desktop get PACKAGE\n");
  fprintf (stderr, "       apt-desktop list\n");
  fprintf (stderr, "       apt-desktop add-interesting PACKAGE...\n");
  fprintf (stderr, "       apt-desktop set-interesting PACKAGE...\n");
  exit (1);
}

void
skip_package (const char *package, const char *version)
{
  dbg ("Skipping %s %s.\n", package, version);
  export_entry (package, NULL);
}

void
update ()
{
  init_state ();
  init_interesting_packages ();
  init_db ();
  init_apt ();

  if (flag_verbose)
    {
      printf ("Updating desktop entries... ");
      fflush (stdout);
    }

  for (pkgCache::PkgIterator pkg = apt_cache->PkgBegin();
       pkg.end() != true;
       pkg++)
    {
      if (should_consider_package (pkg))
        {
          pkgCache::VerIterator inst = installed (pkg);
          if (!inst.end())
            {
              char *old_ver = get_state ("I", pkg.Name());
              append_new_state ("I", pkg.Name(), inst.VerStr());
              if (!old_ver || strcmp (old_ver, inst.VerStr()) != 0)
                {
                  // It is now implicitly interesting, so we remove it
                  // from the list of explicitly interesting packages
                  // if that is requested.
                  //
                  if (is_temporarily_explictly_interesting (pkg.Name()))
                    forget_interesting_package (pkg.Name());

                  if (!update_version (pkg, inst)
                      && !update_version_from_content (pkg, inst))
                    export_entry (pkg.Name(), NULL);
                }
              else
                skip_package (pkg.Name(), inst.VerStr());
              free (old_ver);
            }
          else
            {
              pkgCache::VerIterator cand = candidate (pkg);
              if (!cand.end())
                {
                  char *old_ver = get_state ("C", pkg.Name());
                  append_new_state ("C", pkg.Name(), cand.VerStr());
                  if (!old_ver || strcmp (old_ver, cand.VerStr()) != 0)
                    {
                      if (!update_version (pkg, cand))
                        export_entry (pkg.Name(), NULL);
                    }
                  else
                    skip_package (pkg.Name(), cand.VerStr());
                }
            }
        }
    }

  // XXX - update all explicitly interesting packages here that have
  //       neither a installed nor a candidate version.

  remove_stale_mappings ();

  save_new_state ();
  save_interesting_packages ();

  if (flag_verbose)
    printf ("Done\n");
}

void
get_entry (const char *package)
{
  init_db ();

  char *entry = entry_path (package);
  if (entry)
    printf ("%s\n", entry);
  else
    exit (1);
}

void
list_entries ()
{
  init_db ();
  
  struct dirent *ep;
  DIR *d = must_fdopendir (db_fd);
  while (ep = readdir (d))
    {
      if (strcmp (ep->d_name, ".") == 0
          || strcmp (ep->d_name, "..") == 0)
        continue;

      char *path = xreadlinkat (db_fd, ep->d_name);
      printf ("%s: %s\n", ep->d_name, path);
      free (path);
    }
  must_closedir (d);
}

void
change_package_interest (const char *package, int reason)
{
  if (explicitly_interesting (package) == reason)
    return;

  if (reason != 0)
    {
      pkgCache::PkgIterator pkg = apt_cache->FindPkg (package);
      if (!should_consider_package (pkg))
        return;
    }

  remember_interesting_package (package, reason);
  export_entry (package, NULL);
}

void 
change_packages_interest (char **packages_and_options)
{
  int reason = interest_normal;
  for (char **p = packages_and_options; *p; p++)
    {
      if (strcmp (*p, "--had-error") == 0)
        reason = interest_had_error;
      else if (strcmp (*p, "--normal") == 0)
        reason = interest_normal;
      else
        change_package_interest (*p, reason);
    }
}

void
set_interesting_packages (char **packages_and_options)
{
  init_interesting_packages ();
  init_db ();
  init_apt ();

  change_packages_interest (packages_and_options);

  for (map<string,int>::iterator it = interesting_packages.begin();
       it != interesting_packages.end(); it++)
    {
      char **p;
      for (p = packages_and_options; *p; p++)
        if (it->first == *p)
          break;
      if (*p == NULL)
        change_package_interest (it->first.c_str(), 0);
    }

  save_interesting_packages ();
}

void
add_interesting_packages (char **packages_and_options)
{
  init_interesting_packages ();
  init_db ();
  init_apt ();

  change_packages_interest (packages_and_options);
  save_interesting_packages ();
}

int
main (int argc, char **argv)
{
  // Gobble up options
  while (argc > 1 && argv[1][0] == '-')
    {
      if (argv[1][1] != '-')
        {
          for (const char *p = argv[1]+1; *p; p++)
            {
              if (*p == 'D')
                flag_debug = 1;
              else if (*p == 'v')
                flag_verbose = 1;
              else if (*p == 'a')
                flag_all_packages = 1;
              else
                exitf (1, "Unknown option '%c'.\n", *p);
            }
        }
      else
        exitf (1, "Unknown option '%s'.\n", argv[1]);
        
      argc--;
      argv++;
    }

  /* Debug output interferes with progress reporting.
   */
  if (flag_debug)
    flag_verbose = 0;

  if (argc == 2 && strcmp (argv[1], "update") == 0)
    {
      update ();
    }
  else if (argc == 3 && strcmp (argv[1], "get") == 0)
    {
      get_entry (argv[2]);
    }
  else if (argc == 2 && strcmp (argv[1], "list") == 0)
    {
      list_entries ();
    }
  else if (argc >= 2 && strcmp (argv[1], "set-interesting") == 0)
    {
      set_interesting_packages (argv+2);
    }
  else if (argc >= 2 && strcmp (argv[1], "add-interesting") == 0)
    {
      add_interesting_packages (argv+2);
    }
  else
    usage ();

  exit (0);
}
